/*
 * Copyright (c) 2012 David MASSENOT - http://artcustomer.fr/
 * 
 * Permission is hereby granted to use, modify, and distribute this file 
 * in accordance with the terms of the license agreement accompanying it.
 */

package artcustomer.framework.extensions.away3d.template {
	import flash.geom.Vector3D;
	
	import artcustomer.framework.engines.layers.tools.template.AbstractEngineTemplate;
	import artcustomer.framework.events.InputsEvent;
	
	import away3d.containers.ObjectContainer3D;
	import away3d.containers.View3D;
    import away3d.containers.Scene3D;
	import away3d.cameras.Camera3D;
	import away3d.controllers.HoverController;
	
	
	/**
	 * Away3DTemplate 4.1
	 * 
	 * @author David Massenot
	 */
	public class Away3DTemplate extends AbstractEngineTemplate {
		private var _view:View3D;
		private var _scene:Scene3D;
		private var _camera:Camera3D;
		private var _cameraController:HoverController;
		
		private var _lastPanAngle:Number;
		private var _lastTiltAngle:Number;
		private var _lastMouseX:Number;
		private var _lastMouseY:Number;
		private var _move:Boolean = false;
		private var _lookAtPosition:Vector3D = new Vector3D();
		
		private var _enableController:Boolean = false;
		
		
		/**
		 * Constructor
		 */
		public function Away3DTemplate() {
			super();
		}
		
		//---------------------------------------------------------------------
		//  Events
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function listenEvents():void {
			this.layer3D.context.addEventListener(InputsEvent.INPUT_MOUSE_DOWN, handleInputs, false, 0, true);
			this.layer3D.context.addEventListener(InputsEvent.INPUT_MOUSE_UP, handleInputs, false, 0, true);
			this.layer3D.context.addEventListener(InputsEvent.INPUT_MOUSE_LEAVE, handleInputs, false, 0, true);
		}
		
		/**
		 * @private
		 */
		private function unlistenEvents():void {
			this.layer3D.context.removeEventListener(InputsEvent.INPUT_MOUSE_DOWN, handleInputs);
			this.layer3D.context.removeEventListener(InputsEvent.INPUT_MOUSE_UP, handleInputs);
			this.layer3D.context.removeEventListener(InputsEvent.INPUT_MOUSE_LEAVE, handleInputs);
		}
		
		//---------------------------------------------------------------------
		//  Listeners
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function handleInputs(e:InputsEvent):void {
			switch (e.type) {
				case(InputsEvent.INPUT_MOUSE_DOWN):
					_lastMouseX = this.layer3D.stage.mouseX;
					_lastMouseY = this.layer3D.stage.mouseY;
					_move = true;
					break;
					
				case(InputsEvent.INPUT_MOUSE_UP):
					_move = false;
					break;
					
				case(InputsEvent.INPUT_MOUSE_LEAVE):
					_move = false;
					break;
					
				default:
					break;
			}
		}
		
		//---------------------------------------------------------------------
		//  View3D
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function setupView3D():void {
			_view = new View3D();
			_view.contextMenu = null;
			
			this.layer3D.context.contextView.addChildAt(_view, 0);
			
			_scene = _view.scene;
			_camera = _view.camera;
			_view.antiAlias = 4;
			_camera.lens.far = 5000;
		}
		
		/**
		 * @private
		 */
		private function destroyView3D():void {
			if (_view) {
				this.layer3D.context.contextView.removeChild(_view);
				
				_view.dispose();
				_view = null;
				_scene = null;
				_camera = null;
			}
		}
		
		/**
		 * @private
		 */
		private function renderView3D():void {
			if (_view) _view.render();
		}
		
		/**
		 * @private
		 */
		private function resizeView3D():void {
			if (_view) {
				_view.width = this.layer3D.context.contextWidth;
				_view.height = this.layer3D.context.contextHeight;
			}
		}
		
		//---------------------------------------------------------------------
		//  Controller
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function setupController():void {
			_cameraController = new HoverController(_view.camera, null, 45, 20, 2000, 5);
		}
		
		/**
		 * @private
		 */
		private function destroyController():void {
			_cameraController = null;
		}
		
		/**
		 * @private
		 */
		private function renderController():void {
			if (_cameraController) {
				if (_move) {
					_cameraController.panAngle = 0.3 * (this.layer3D.stage.mouseX - _lastMouseX) + _lastPanAngle;
					_cameraController.tiltAngle = 0.3 * (this.layer3D.stage.mouseY - _lastMouseY) + _lastTiltAngle;
				} else {
					_lastPanAngle = _cameraController.panAngle;
					_lastTiltAngle = _cameraController.tiltAngle;
				}
				
				_cameraController.lookAtPosition = _lookAtPosition;
			}
		}
		
		
		/**
		 * Setup template.
		 */
		override public function setup():void {
			super.setup();
			
			setupView3D();
			setupController();
			
			this.setController(true);
			
			this.resize();
		}
		
		/**
		 * Destroy template.
		 */
		override public function destroy():void {
			this.setController(false);
			
			unlistenEvents();
			destroyController();
			destroyView3D();
			
			super.destroy();
		}
		
		/**
		 * Render template.
		 */
		override public function render():void {
			renderView3D();
			renderController();
		}
		
		/**
		 * Resize template.
		 */
		override public function resize():void {
			resizeView3D();
		}
		
		/**
		 * Add object in scene.
		 * @param	object
		 * @return
		 */
		public function addObject(object:ObjectContainer3D):ObjectContainer3D {
			if (_view) return _view.scene.addChild(object);
			
			return null;
		}
		
		/**
		 * Remove object in scene.
		 * @param	object
		 */
		public function removeObject(object:ObjectContainer3D):void {
			if (_view) _view.scene.removeChild(object);
		}
		
		/**
		 * Enable / Disable camera controller.
		 * 
		 * @param	value
		 */
		public function setController(value:Boolean):void {
			if (_enableController == value) return;
			
			_enableController = value;
			
			if (_enableController) {
				listenEvents();
			} else {
				unlistenEvents();
			}
		}
		
		
		/**
		 * @private
		 */
		public function get view():View3D {
			return _view;
		}
		
		/**
		 * @private
		 */
		public function get scene():Scene3D {
			return _scene;
		}
		
		/**
		 * @private
		 */
		public function get camera():Camera3D {
			return _camera;
		}
		
		/**
		 * @private
		 */
		public function get cameraController():HoverController {
			return _cameraController;
		}
		
		/**
		 * @private
		 */
		public function get enableController():Boolean {
			return _enableController;
		}
	}
}