/*
 * Copyright (c) 2013 David MASSENOT - http://artcustomer.fr/
 * 
 * Permission is hereby granted to use, modify, and distribute this file 
 * in accordance with the terms of the license agreement accompanying it.
 */

package artcustomer.framework.extensions.display.ui.list {
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.utils.getQualifiedClassName;
	
	import artcustomer.framework.base.IDestroyable;
	import artcustomer.framework.errors.IllegalError;
	import artcustomer.framework.extensions.display.ui.list.events.MenuListEvent;
	import artcustomer.framework.extensions.display.ui.list.errors.AbstractListError;
	import artcustomer.framework.extensions.display.ui.list.data.*;
	import artcustomer.framework.extensions.display.ui.list.layout.*;
	import artcustomer.framework.extensions.display.ui.list.item.*;
	
	[Event(name = "onSelectItem", type = "artcustomer.framework.extensions.display.ui.list.events.MenuListEvent")]
	[Event(name = "onFocusItem", type = "artcustomer.framework.extensions.display.ui.list.events.MenuListEvent")]
	[Event(name = "onLockItem", type = "artcustomer.framework.extensions.display.ui.list.events.MenuListEvent")]
	[Event(name = "onUnlockItem", type = "artcustomer.framework.extensions.display.ui.list.events.MenuListEvent")]
	[Event(name = "onUnfocusItem", type = "artcustomer.framework.extensions.display.ui.list.events.MenuListEvent")]
	[Event(name = "onListChange", type = "artcustomer.framework.extensions.display.ui.list.events.MenuListEvent")]
	
	
	/**
	 * AbstractList
	 * 
	 * @author David Massenot
	 */
	public class AbstractList extends Sprite implements IDestroyable {
		private static const FULL_CLASS_NAME:String = 'artcustomer.maxima.extensions.display.ui.list::AbstractList';
		
		private var _container:Sprite;
		private var _items:Vector.<AbstractListItem>;
		
		private var _layout:IListLayout;
		
		private var _currentItem:AbstractListItem;
		private var _lockedItem:AbstractListItem;
		private var _numItems:int;
		
		private var _iteratorIndex:int;
		
		
		/**
		 * Constructor
		 */
		public function AbstractList() {
			if (getQualifiedClassName(this) == FULL_CLASS_NAME) throw new IllegalError(IllegalError.E_ABSTRACT_CLASS);
			
			init();
			setupContainer();
			createStack();
		}
		
		//---------------------------------------------------------------------
		//  Initialize
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function init():void {
			_iteratorIndex = 0;
		}
		
		//---------------------------------------------------------------------
		//  Container
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function setupContainer():void {
			_container = new Sprite();
			
			this.addChild(_container);
		}
		
		/**
		 * @private
		 */
		private function destroyContainer():void {
			this.removeChild(_container);
			
			_container = null;
		}
		
		//---------------------------------------------------------------------
		//  Stack
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function createStack():void {
			_items = new Vector.<AbstractListItem>();
		}
		
		/**
		 * @private
		 */
		private function releaseStack():void {
			while (_items.length > 0) {
				_items.shift().destroy();
			}
		}
		
		/**
		 * @private
		 */
		private function destroyStack():void {
			_items = null;
		}
		
		//---------------------------------------------------------------------
		//  Items
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function updateItemIndexes():void {
			var i:int = 0;
			var item:AbstractListItem;
			
			while (i < _items.length) {
				item = _items[i];
				item.valueObject.index = i;
				
				i++;
			}
		}
		
		/**
		 * @private
		 */
		private function removeAllItems():void {
			var i:int = 0;
			var item:AbstractListItem;
			
			while (i < _items.length) {
				item = _items[i];
				
				_container.removeChild(item);
				
				i++;
			}
		}
		
		/**
		 * @private
		 */
		private function toggleCurrentItem(item:AbstractListItem):void {
			var lastItemIndex:int = 0;
			var currentItemIndex:int;
			
			if (_currentItem) {
				lastItemIndex = _currentItem.valueObject.index;
				
				_currentItem.unfocus();
				_currentItem = null;
			}
			
			if (item) {
				_currentItem = item;
				_currentItem.focus();
				
				currentItemIndex = item.valueObject.index;
				
				onItemFocused(currentItemIndex >= lastItemIndex);
			}
		}
		
		/**
		 * @private
		 */
		private function toggleLockedItem(item:AbstractListItem):void {
			if (_lockedItem) {
				_lockedItem.unlock();
				_lockedItem = null;
			}
			
			if (item) {
				_lockedItem = item;
				_lockedItem.lock();
			}
		}
		
		//---------------------------------------------------------------------
		//  Layout
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function displayLayout():void {
			if (_layout) _layout.display();
		}
		
		/**
		 * @private
		 */
		private function destroyLayout():void {
			if (_layout) {
				(_layout as AbstractListLayout).destroy();
				_layout = null;
			}
		}
		
		/**
		 * @private
		 */
		private function dispatchMenuListEvent(type:String):void {
			this.dispatchEvent(new MenuListEvent(type, false, false));
		}
		
		
		/**
		 * Call when item is added. Override it !
		 */
		protected function onItemAdded():void {
			
		}
		
		/**
		 * Call when item is focused. Override it !
		 * 
		 * @param	next
		 */
		protected function onItemFocused(next:Boolean):void {
			
		}
		
		
		/**
		 * Destructor.
		 */
		public function destroy():void {
			removeAllItems();
			destroyContainer();
			releaseStack();
			destroyStack();
			
			_currentItem = null;
			_lockedItem = null;
			_numItems = 0;
			_iteratorIndex = 0;
		}
		
		/**
		 * Add item at index.
		 * 
		 * @param	itemClass
		 * @param	valueObject
		 * @param	index
		 */
		public function addItemAt(itemClass:Class, valueObject:AbstractListItemValueObject, index:int = -1):void {
			if (!itemClass is AbstractListItem) throw new AbstractListError(AbstractListError.E_LIST_ADDITEM_SUPER);
			
			var item:AbstractListItem = new itemClass();
			
			if (item) {
				item.valueObject = valueObject;
				
				if (index > -1) {
					_items.splice(index, 0, item);
					
					if (valueObject) {
						item.valueObject.index = index;
						
						updateItemIndexes();
					}
					
					displayLayout();
				} else {
					if (valueObject) item.valueObject.index = _numItems;
					
					_items.push(item);
				}
				
				_container.addChild(item);
				
				item.setup();
				
				_numItems++;
				
				onItemAdded();
				dispatchMenuListEvent(MenuListEvent.ON_LIST_CHANGE);
			} else {
				throw new AbstractListError(AbstractListError.E_LIST_ADDITEM);
			}
		}
		
		/**
		 * Remove item at index.
		 * 
		 * @param	index
		 */
		public function removeItemAt(index:int):Boolean {
			var item:AbstractListItem = this.getItemAt(index);
			
			if (item) {
				item = _items.shift();
				item.destroy();
				
				_container.removeChild(item);
				
				updateItemIndexes();
				displayLayout();
				
				_numItems--;
				
				dispatchMenuListEvent(MenuListEvent.ON_LIST_CHANGE);
				
				return true;
			}
			
			return false;
		}
		
		/**
		 * Get item at index.
		 * 
		 * @param	index
		 * @return
		 */
		public function getItemAt(index:int):AbstractListItem {
			if (this.hasItemAt(index)) return _items[index];
			return null;
		}
		
		/**
		 * Test item at index.
		 * 
		 * @param	index
		 * @return
		 */
		public function hasItemAt(index:int):Boolean {
			return index >= 0 && index < _numItems;
		}
		
		/**
		 * Remove all items.
		 */
		public function removeAll():void {
			while (_numItems > 0) {
				this.removeItemAt(_numItems - 1);
			}
			
			_currentItem = null;
			_lockedItem = null;
			_numItems = 0;
			_iteratorIndex = 0;
			
			dispatchMenuListEvent(MenuListEvent.ON_LIST_CLEARED);
		}
		
		/**
		 * Test next move.
		 * 
		 * @return
		 */
		public function hasNext():Boolean {
			return _iteratorIndex < _numItems - 1;
		}
		
		/**
		 * Test previous move.
		 * 
		 * @return
		 */
		public function hasPrevious():Boolean {
			return _iteratorIndex > 0;
		}
		
		/**
		 * Select current item.
		 */
		public function select():void {
			if (_currentItem) _currentItem.select();
		}
		
		/**
		 * Unselect current item.
		 */
		public function unselect():void {
			toggleCurrentItem(null);
			toggleLockedItem(null);
		}
		
		/**
		 * Focus item at index.
		 * 
		 * @param	index
		 */
		public function focusAt(index:int):void {
			var item:AbstractListItem = this.getItemAt(index);
			
			if (item) {
				_iteratorIndex = index;
				
				toggleCurrentItem(item);
			}
		}
		
		/**
		 * Focus the next item.
		 */
		public function focusNext():void {
			var item:AbstractListItem;
			
			_iteratorIndex++;
			
			if (_iteratorIndex >= _numItems) _iteratorIndex = 0;
			
			item = this.getItemAt(_iteratorIndex);
			
			if (item) toggleCurrentItem(item);
		}
		
		/**
		 * Focus the previous item.
		 */
		public function focusPrevious():void {
			var item:AbstractListItem;
			
			_iteratorIndex--;
			
			if (_iteratorIndex < 0) _iteratorIndex = _numItems - 1;
			
			item = this.getItemAt(_iteratorIndex);
			
			if (item) toggleCurrentItem(item);
		}
		
		/**
		 * Unfocus selected item.
		 */
		public function unfocus():void {
			if (_currentItem) _currentItem.unfocus();
		}
		
		/**
		 * Unlocked locked item.
		 */
		public function unlock():void {
			if (_lockedItem) _lockedItem.unlock();
		}
		
		/**
		 * Unlocked locked item.
		 */
		public function lockAt(index:int):void {
			var item:AbstractListItem = this.getItemAt(index);
			
			if (item) toggleLockedItem(item);
		}
		
		/**
		 * Set Layout.
		 * 
		 * @param	layout
		 */
		public function setLayout(layout:IListLayout):void {
			if (!layout is AbstractListLayout) throw new AbstractListError(AbstractListError.E_LIST_SETLAYOUT_SUPER);
			
			if (layout) {
				destroyLayout();
				
				_layout = layout;
				(_layout as AbstractListLayout).displayItems = _items;
				
				displayLayout();
			} else {
				throw new AbstractListError(AbstractListError.E_LIST_SETLAYOUT);
			}
		}
		
		
		/**
		 * @private
		 */
		protected function get container():Sprite {
			return _container;
		}
		
		/**
		 * @private
		 */
		public function get layout():IListLayout {
			return _layout;
		}
		
		/**
		 * @private
		 */
		public function get numItems():int {
			return _numItems;
		}
		
		/**
		 * @private
		 */
		public function get iteratorIndex():int {
			return _iteratorIndex;
		}
		
		/**
		 * @private
		 */
		public function get currentItem():AbstractListItem {
			return _currentItem;
		}
	}
}