/*
 * Copyright (c) 2013 David MASSENOT - http://artcustomer.fr/
 * 
 * Permission is hereby granted to use, modify, and distribute this file 
 * in accordance with the terms of the license agreement accompanying it.
 */

package artcustomer.framework.extensions.display.ui.list {
	import flash.display.Shape;
	import flash.display.Sprite;
	
	import artcustomer.framework.extensions.display.ui.list.item.AbstractListItem;
	import artcustomer.framework.extensions.display.ui.list.layout.VerticalListLayout;
	import artcustomer.framework.extensions.display.ui.list.scroll.VerticalScrollBar;
	
	
	/**
	 * KeyScrollableList
	 * 
	 * @author David Massenot
	 */
	public class KeyScrollableList extends AbstractList {
		private var _mask:Shape;
		private var _layout:VerticalListLayout;
		private var _scrollBar:VerticalScrollBar;
		
		private var _itemWidth:int;
		private var _itemHeight:int;
		
		private var _listStep:int;
		private var _numVisibleItems:int;
		
		private var _isMask:Boolean;
		private var _isScroll:Boolean;
		
		
		/**
		 * Constructor
		 */
		public function KeyScrollableList() {
			super();
			
			setDisplayLayout();
			setupMask();
			setupScroll();
		}
		
		//---------------------------------------------------------------------
		//  Container
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function resetContainer():void {
			this.container.y = 0;
		}
		
		//---------------------------------------------------------------------
		//  Layout
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function setDisplayLayout():void {
			_layout = new VerticalListLayout();
			_layout.margin = 5;
			
			this.setLayout(_layout);
		}
		
		//---------------------------------------------------------------------
		//  Mask
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function setupMask():void {
			_mask = new Shape();
			
			this.addChild(_mask);
		}
		
		/**
		 * @private
		 */
		private function destroyMask():void {
			this.removeChild(_mask);
			
			_mask = null;
		}
		
		/**
		 * @private
		 */
		private function drawMask():void {
			if (this.numItems != 0) {
				var itemReference:AbstractListItem = this.getItemAt(0);
				var maskHeight:int;
				
				_itemWidth = itemReference.width;
				_itemHeight = itemReference.height;
				
				maskHeight = (_numVisibleItems * (_itemHeight + _layout.margin)) - _layout.margin;
				
				_mask.graphics.clear();
				_mask.graphics.beginFill(0x000000, 1);
				_mask.graphics.drawRect(0, 0, _itemWidth, maskHeight);
				_mask.graphics.endFill();
			}
		}
		
		/**
		 * @private
		 */
		private function clearMask():void {
			_mask.graphics.clear();
		}
		
		/**
		 * @private
		 */
		private function setMask():void {
			this.container.mask = _mask
		}
		
		/**
		 * @private
		 */
		private function unsetMask():void {
			this.container.mask = null;
		}
		
		//---------------------------------------------------------------------
		//  Scroll
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function setupScroll():void {
			_scrollBar = new VerticalScrollBar();
			
			this.addChild(_scrollBar);
		}
		
		/**
		 * @private
		 */
		private function destroyScroll():void {
			this.removeChild(_scrollBar);
			
			_scrollBar = null;
		}
		
		/**
		 * @private
		 */
		private function drawScroll():void {
			var itemReference:AbstractListItem = this.getItemAt(0);
			var areaHeight:int = _mask.height;
			var listHeight:int = ((itemReference.height + _layout.margin) * this.numItems) - _layout.margin;
			var size:int = areaHeight / listHeight * areaHeight;
			
			_scrollBar.draw(size);
		}
		
		/**
		 * @private
		 */
		private function clearScroll():void {
			_scrollBar.clear();
		}
		
		/**
		 * @private
		 */
		private function updateScroll():void {
			var itemReference:AbstractListItem = this.getItemAt(0);
			var areaHeight:int = _mask.height;
			var listHeight:int = ((itemReference.height + _layout.margin) * this.numItems) - _layout.margin;
			
			_scrollBar.y = areaHeight * Math.min(1, (-this.container.y / listHeight));
		}
		
		/**
		 * @private
		 */
		private function moveScroll():void {
			_scrollBar.x = _mask.x + _mask.width + 5;
			_scrollBar.y = 0;
		}
		
		
		/**
		 * Call when item is added.
		 */
		override protected function onItemAdded():void {
			if (this.numItems > _numVisibleItems) {
				if (!_isMask) {
					_isMask = true;
					_isScroll = true;
					
					drawMask();
					setMask();
					moveScroll();
				}
				
				if (_isScroll) {
					drawScroll();
					updateScroll();
				}
			} else {
				if (_isMask) {
					resetContainer();
					clearScroll();
					clearMask();
					unsetMask();
					
					_isMask = false;
					_isScroll = false;
				}
			}
		}
		
		/**
		 * Call when item is focused.
		 * 
		 * @param	next
		 */
		override protected function onItemFocused(next:Boolean):void {
			if (_isMask) {
				if (next) {
					if (this.currentItem.y + this.container.y >= _mask.y + _mask.height) {
						_listStep = this.iteratorIndex - _numVisibleItems + 1;
						
						this.container.y = (_itemHeight + _layout.margin) * -_listStep;
					}
				} else {
					if (this.currentItem.y + this.container.y < _mask.y) {
						_listStep = this.iteratorIndex;
						
						this.container.y = (_itemHeight + _layout.margin) * -_listStep;
					}
				}
				
				updateScroll();
			}
		}
		
		/**
		 * Focus the next item.
		 */
		override public function focusNext():void {
			if (this.hasNext()) super.focusNext();
		}
		
		
		/**
		 * Focus the previous item.
		 */
		override public function focusPrevious():void {
			if (this.hasPrevious()) super.focusPrevious();
		}
		
		/**
		 * Remove all items.
		 */
		override public function removeAll():void {
			resetContainer();
			clearMask();
			unsetMask();
			clearScroll();
			
			_listStep = 0;
			_isMask = false;
			_isScroll = false;
			
			super.removeAll();
		}
		
		/**
		 * Destructor
		 */
		override public function destroy():void {
			destroyScroll();
			destroyMask();
			
			_layout = null;
			_listStep = 0;
			_numVisibleItems = 0;
			_itemWidth = 0;
			_itemHeight = 0;
			_isMask = false;
			_isScroll = false;
			
			super.destroy();
		}
		
		
		/**
		 * @private
		 */
		public function get numVisibleItems():Number {
			return _numVisibleItems;
		}
		
		/**
		 * @private
		 */
		public function set numVisibleItems(value:Number):void {
			_numVisibleItems = value;
			
			drawMask();
		}
	}
}