/*
 * Copyright (c) 2012 David MASSENOT - http://artcustomer.fr/
 * 
 * Permission is hereby granted to use, modify, and distribute this file 
 * in accordance with the terms of the license agreement accompanying it.
 */

package artcustomer.framework.extensions.starling.layers {
	import flash.geom.Rectangle;
	import flash.utils.getQualifiedClassName;
	
	import artcustomer.framework.engines.layers.Layer3D;
	import artcustomer.framework.context.root.StarlingRootClass;
	import artcustomer.framework.extensions.starling.events.LayerStarlingEvent;
	import artcustomer.framework.extensions.starling.errors.FrameworkStarlingError;
	import artcustomer.framework.extensions.starling.layers.tools.scene.AbstractStarlingScene;
	
	import starling.core.Starling;
	import starling.display.Stage;
	import starling.events.Event;
	import starling.events.ResizeEvent;
	
	[Event(name = "starlingContextSetup", type = "artcustomer.framework.extensions.starling.events.LayerStarlingEvent")]
	[Event(name = "starlingContextDestroy", type = "artcustomer.framework.extensions.starling.events.LayerStarlingEvent")]
	[Event(name = "starlingContextError", type = "artcustomer.framework.extensions.starling.events.LayerStarlingEvent")]
	[Event(name = "starlingContextStart", type = "artcustomer.framework.extensions.starling.events.LayerStarlingEvent")]
	
	
	/**
	 * LayerStarling
	 * 
	 * @author David Massenot
	 */
	public class LayerStarling extends Layer3D {
		private static const FULL_CLASS_NAME:String = 'artcustomer.framework.extensions.starling.layers::LayerStarling';
		
		private var _starling:Starling;
		private var _starlingRoot:StarlingRootClass;
		
		private var _starlingStage:Stage;
		
		private var _handleLostContext:Boolean;
		
		
		/**
		 * Constructor
		 */
		public function LayerStarling() {
			super();
			
			if (getQualifiedClassName(this) == FULL_CLASS_NAME) throw new FrameworkStarlingError(FrameworkStarlingError.E_STARLINGCONTEXT_CONSTRUCTOR);
		}
		
		//---------------------------------------------------------------------
		//  Starling
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function setupStarling():void {
			var viewPort:Rectangle = new Rectangle(0, 0, this.context.contextWidth, this.context.contextHeight);
			
			try {
				_starling = new Starling(StarlingRootClass, this.stage, viewPort, this.stage3D);
				
				_starling.addEventListener(Event.ROOT_CREATED, handleStarling);
				_starling.stage.addEventListener(ResizeEvent.RESIZE, handleStarlingResize);
				
				_starling.shareContext = true;
				_starling.enableErrorChecking = false;
				_starling.start();
				
				_starlingStage = _starling.stage;
			} catch (er:Error) {
				dispatchStarlingEvent(LayerStarlingEvent.STARLINGCONTEXT_ERROR, er.message);
			}
		}
		
		/**
		 * @private
		 */
		private function destroyStarling():void {
			if (_starling) {
				_starling.removeEventListener(Event.ROOT_CREATED, handleStarling);
				_starling.stage.removeEventListener(ResizeEvent.RESIZE, handleStarlingResize);
				_starling.dispose();
				_starling = null;
			}
		}
		
		/**
		 * @private
		 */
		private function resizeStarling():void {
			var width:int = this.context.contextWidth;
			var height:int = this.context.contextHeight;
			var viewport:Rectangle = new Rectangle(0, 0, width, height);
			
			if (_starling) {
				_starling.viewPort = viewport;
				_starling.stage.stageWidth = width;
				_starling.stage.stageHeight= height;
			}
		}
		
		/**
		 * @private
		 */
		private function renderStarling():void {
			if (_starling) _starling.nextFrame();
		}
		
		/**
		 * @private
		 */
		private function handleStarling(e:Event):void {
			switch (e.type) {
				case(Event.ROOT_CREATED):
					if (_starling) _starlingRoot = _starling.root as StarlingRootClass;
					
					start();
					dispatchStarlingEvent(LayerStarlingEvent.STARLINGCONTEXT_START);
					break;
					
				default:
					break;
			}
		}
		
		/**
		 * @private
		 */
		private function handleStarlingResize(e:ResizeEvent):void {
			switch (e.type) {
				case(ResizeEvent.RESIZE):
					resizeStarling();
					break;
					
				default:
					break;
			}
		}
		
		//---------------------------------------------------------------------
		//  Event Dispatching
		//---------------------------------------------------------------------
		
		/**
		 * @private
		 */
		private function dispatchStarlingEvent(type:String, error:String = null):void {
			this.context.dispatchEvent(new LayerStarlingEvent(type, false, false, this, this.stage3D, error));
		}
		
		
		/**
		 * Setup StarlingContext.
		 */
		override public function setup():void {
			super.setup();
			
			_handleLostContext = false;
			
			dispatchStarlingEvent(LayerStarlingEvent.STARLINGCONTEXT_SETUP);
		}
		
		/**
		 * Destroy StarlingContext.
		 */
		override public function destroy():void {
			dispatchStarlingEvent(LayerStarlingEvent.STARLINGCONTEXT_DESTROY);
			destroyStarling();
			
			_starling = null;
			_starlingRoot = null;
			_starlingStage = null;
			_handleLostContext = false;
			
			super.destroy();
		}
		
		/**
		 * Called when Context3D is created.
		 */
		override protected function create():void {
			setupStarling();
		}
		
		/**
		 * Called when Starling is resized. Override it !
		 */
		override protected function resize():void {
			var id:String;
			var scene:AbstractStarlingScene;
			
			for (id in scenes) {
				scene = scenes[id] as AbstractStarlingScene;
				if (scene) scene.resize();
			}
		}
		
		/**
		 * Called when RenderEngine is on.
		 */
		override protected function render():void {
			renderStarling();
			
			this.context3D.present();
		}
		
		/**
		 * Called when Starling is started. Override it !
		 */
		protected function start():void {
			
		}
		
		/**
		 * Add new AbstractStarlingScene.
		 * 
		 * @param	sceneClass : Class of the instance of AbstractStarlingScene.
		 * @return ths instance created.
		 */
		public final function addDisplayScene(sceneClass:Class):AbstractStarlingScene {
			var scene:AbstractStarlingScene = this.addScene(sceneClass) as AbstractStarlingScene;
			
			scene.rootClass = _starlingRoot;
			scene.start();
			
			return scene;
		}
		
		
		/**
		 * @private
		 */
		public function get starling():Starling {
			return _starling;
		}
		
		/**
		 * @private
		 */
		public function get starlingRoot():StarlingRootClass {
			return _starlingRoot;
		}
		
		/**
		 * @private
		 */
		public function get starlingStage():Stage {
			return _starlingStage;
		}
		
		/**
		 * @private
		 */
		public function get handleLostContext():Boolean {
			return _handleLostContext;
		}
		
		/**
		 * @private
		 */
		public function set handleLostContext(value:Boolean):void {
			if (value != _handleLostContext) {
				_handleLostContext = value;
				
				Starling.handleLostContext = _handleLostContext;
			}
		}
	}
}